@@ -1,7 +1,7 @@
 /*
  * yamdi.c
  *
- * Copyright (c) 2007, Ingo Oppermann
+ * Copyright (c) 2007-2008, Ingo Oppermann
  * All rights reserved.
  * 
  * Redistribution and use in source and binary forms, with or without
@@ -36,7 +36,11 @@
  */
 
 #include <sys/types.h>
-#include <sys/mman.h>
+#ifdef __MINGW32__
+	#include <windows.h>
+#else
+	#include <sys/mman.h>
+#endif
 #include <sys/stat.h>
 #include <stdio.h>
 #include <stdlib.h>
@@ -58,7 +62,7 @@
 #define	FLV_VP6ALPHAVIDEOPACKET	5
 #define FLV_SCREENV2VIDEOPACKET	6
 
-#define YAMDI_VERSION	"1.2"
+#define YAMDI_VERSION	"1.4"
 
 #ifndef MAP_NOCORE
 	#define MAP_NOCORE	0
@@ -141,7 +145,7 @@ typedef struct {
 
 void initFLVMetaData(const char *creator, int lastsecond, int lastkeyframe);
 size_t writeFLVMetaData(FILE *fp);
-size_t writeFLVLastSecond(FILE *fp);
+size_t writeFLVLastSecond(FILE *fp, double timestamp);
 size_t writeFLVLastKeyframe(FILE *fp);
 
 void readFLVFirstPass(char *flv, size_t streampos, size_t filesize);
@@ -168,14 +172,18 @@ size_t writeFLVBool(FILE *fp, int value);
 size_t writeFLVDouble(FILE *fp, double v);
 
 void writeFLV(FILE *fp, char *flv, size_t streampos, size_t filesize);
+void writeXMLMetadata(FILE *fp, const char *infile, const char *outfile);
 void writeFLVHeader(FILE *fp);
 
 void print_usage(void);
 
 int main(int argc, char *argv[]) {
-	FILE *fp_infile = NULL, *fp_outfile = NULL, *devnull;
+	FILE *fp_infile = NULL, *fp_outfile = NULL, *fp_xmloutfile = NULL, *devnull;
+#ifdef __MINGW32__
+	HANDLE fh_infile = NULL;
+#endif
 	int c, lastsecond = 0, lastkeyframe = 0;
-	char *flv, *infile, *outfile, *creator;
+	char *flv, *infile, *outfile, *xmloutfile, *creator;
 	unsigned int i;
 	size_t filesize = 0, streampos, metadatasize;
 	struct stat sb;
@@ -185,9 +193,10 @@ int main(int argc, char *argv[]) {
 
 	infile = NULL;
 	outfile = NULL;
+	xmloutfile = NULL;
 	creator = NULL;
 
-	while((c = getopt(argc, argv, "i:o:c:lkh")) != -1) {
+	while((c = getopt(argc, argv, "i:o:x:c:lkh")) != -1) {
 		switch(c) {
 			case 'i':
 				if(infile == NULL) {
@@ -207,7 +216,14 @@ int main(int argc, char *argv[]) {
 
 					filesize = sb.st_size;
 
+#ifndef __MINGW32__
 					fp_infile = fopen(infile, "rb");
+#else
+					// Open infile with CreateFile() API
+					fh_infile = CreateFile(infile, GENERIC_READ, 0, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
+					// Meaningless type casting here. It is just used to pass the error checking codes.
+					fp_infile = (FILE *)fh_infile;
+#endif
 					if(fp_infile == NULL) {
 						fprintf(stderr, "Couldn't open %s.\n", infile);
 						exit(1);
@@ -237,6 +253,17 @@ int main(int argc, char *argv[]) {
 						fp_outfile = stdout;
 				}
 				break;
+			case 'x':
+				if(xmloutfile == NULL) {
+					xmloutfile = optarg;
+		
+					fp_xmloutfile = fopen(xmloutfile, "wb");
+					if(fp_xmloutfile == NULL) {
+						fprintf(stderr, "Couldn't open %s.\n", outfile);
+						exit(1);
+					}
+				}
+				break;
 			case 'c':
 				creator = optarg;
 				break;
@@ -265,13 +292,28 @@ int main(int argc, char *argv[]) {
 		}
 	}
 
-	if(fp_infile == NULL || fp_outfile == NULL) {
-		fprintf(stderr, "Please provide an input file and an output file. -h for help.\n");
+	if(infile == NULL) {
+		fprintf(stderr, "Please provide an input file. -h for help.\n");
+		exit(1);
+	}
+
+	if(outfile == NULL && xmloutfile == NULL) {
+		fprintf(stderr, "Please provide at least one output file. -h for help.\n");
 		exit(1);
 	}
 
 	// mmap von infile erstellen
+#ifndef __MINGW32__
 	flv = mmap(NULL, filesize, PROT_READ, MAP_NOCORE | MAP_PRIVATE, fileno(fp_infile), 0);
+#else
+	HANDLE h = NULL;
+	h = CreateFileMapping(fh_infile, NULL, PAGE_READONLY | SEC_COMMIT, 0, filesize,  NULL);
+	if(h == NULL) {
+		fprintf(stderr, "Couldn't create file mapping object %s. Error code: %d\n", infile, (int)GetLastError());
+		exit(1);
+	}
+	flv = MapViewOfFile(h, FILE_MAP_READ, 0, 0, filesize);
+#endif
 	if(flv == NULL) {
 		fprintf(stderr, "Couldn't load %s.\n", infile);
 		exit(1);
@@ -293,7 +335,12 @@ int main(int argc, char *argv[]) {
 	// Das FLV einlesen und Informationen fuer die Metatags extrahieren
 	readFLVFirstPass(flv, streampos, filesize);
 
+#ifndef __MINGW32__
 	devnull = fopen("/dev/null", "wb");
+#else
+	devnull = fopen("nul", "wb");
+#endif
+
 	if(devnull == NULL) {
 		fprintf(stderr, "Couldn't open NULL device.\n");
 		exit(1);
@@ -301,7 +348,7 @@ int main(int argc, char *argv[]) {
 
 	// Die Groessen berechnen
 	metadatasize = writeFLVMetaData(devnull);
-	flvmetadata.lastsecondsize = writeFLVLastSecond(devnull);
+	flvmetadata.lastsecondsize = writeFLVLastSecond(devnull, 0.0);
 	flvmetadata.lastkeyframesize = writeFLVLastKeyframe(devnull);	// Not fully implemented, i.e. has no effect
 
 	fclose(devnull);
@@ -322,13 +369,18 @@ int main(int argc, char *argv[]) {
 	if(flvmetadata.hasLastSecond == 1)
 		flvmetadata.filesize += (double)flvmetadata.lastsecondsize;
 
-	writeFLV(fp_outfile, flv, streampos, filesize);
+	if(outfile != NULL)
+		writeFLV(fp_outfile, flv, streampos, filesize);
+
+	if(xmloutfile != NULL)
+		writeXMLMetadata(fp_xmloutfile, infile, outfile);
 
 	return 0;
 }
 
 void writeFLV(FILE *fp, char *flv, size_t streampos, size_t filesize) {
 	int tagcount;
+	double currenttimestamp;
 	size_t datasize;
 	FLVTag_t *flvtag;
 
@@ -355,8 +407,10 @@ void writeFLV(FILE *fp, char *flv, size_t streampos, size_t filesize) {
 			if(flvtag->type == FLV_VIDEODATA && flvmetadata.hasLastSecond == 1) {
 				tagcount++;
 
-				if(tagcount == flvmetadata.lastsecondTagCount)
-					writeFLVLastSecond(fp);
+				if(tagcount == flvmetadata.lastsecondTagCount) {
+					currenttimestamp = (double)((flvtag->timestamp_ex << 24) + (flvtag->timestamp[0] << 16) + (flvtag->timestamp[1] << 8) + flvtag->timestamp[2]) / 1000.0;
+					writeFLVLastSecond(fp, currenttimestamp);
+				}
 			}
 		}
 
@@ -366,13 +420,88 @@ void writeFLV(FILE *fp, char *flv, size_t streampos, size_t filesize) {
 	return;
 }
 
+void writeXMLMetadata(FILE *fp, const char *infile, const char *outfile) {
+	int i;
+	char *hasit;
+
+	fprintf(fp, "<?xml version='1.0' encoding='UTF-8'?>\n");
+	fprintf(fp, "<fileset>\n");
+
+	if(outfile != NULL)
+		fprintf(fp, "<flv name=\"%s\">\n", outfile);
+	else
+		fprintf(fp, "<flv name=\"%s\">\n", infile);
+
+	hasit = (flvmetadata.hasKeyframes > 0) ? "true" : "false";
+	fprintf(fp, "<hasKeyframes>%s</hasKeyframes>\n", hasit);		      
+
+	hasit = (flvmetadata.hasVideo > 0) ? "true" : "false";
+	fprintf(fp, "<hasVideo>%s</hasVideo>\n", hasit);		      
+
+	hasit = (flvmetadata.hasAudio > 0) ? "true" : "false";
+	fprintf(fp, "<hasAudio>%s</hasAudio>\n", hasit);		      
+
+	hasit = (flvmetadata.hasMetadata > 0) ? "true" : "false";
+	fprintf(fp, "<hasMetadata>%s</hasMetadata>\n", hasit);		      
+
+	hasit = (flvmetadata.hasCuePoints > 0) ? "true" : "false";
+	fprintf(fp, "<hasCuePoints>%s</hasCuePoints>\n", hasit);		      
+
+	hasit = (flvmetadata.canSeekToEnd > 0) ? "true" : "false";
+	fprintf(fp, "<canSeekToEnd>%s</canSeekToEnd>\n", hasit);		      
+
+	fprintf(fp, "<audiocodecid>%i</audiocodecid>\n", (int)flvmetadata.audiocodecid);		         		      
+	fprintf(fp, "<audiosamplerate>%i</audiosamplerate>\n", (int)flvmetadata.audiosamplerate);
+	fprintf(fp, "<audiodatarate>%i</audiodatarate>\n", (int)flvmetadata.audiodatarate);
+	fprintf(fp, "<audiosamplesize>%i</audiosamplesize>\n", (int)flvmetadata.audiosamplesize);
+	fprintf(fp, "<audiodelay>%.2f</audiodelay>\n", flvmetadata.audiodelay);
+	hasit = (flvmetadata.stereo > 0) ? "true" : "false";
+	fprintf(fp, "<stereo>%s</stereo>\n", hasit);		      
+
+	fprintf(fp, "<videocodecid>%i</videocodecid>\n", (int)flvmetadata.videocodecid);
+	fprintf(fp, "<framerate>%.2f</framerate>\n", flvmetadata.framerate);
+	fprintf(fp, "<videodatarate>%i</videodatarate>\n", (int)flvmetadata.videodatarate);
+	fprintf(fp, "<height>%i</height>\n", (int)flvmetadata.height);
+	fprintf(fp, "<width>%i</width>\n", (int)flvmetadata.width);
+
+	fprintf(fp, "<datasize>%i</datasize>\n", (int)flvmetadata.datasize);
+	fprintf(fp, "<audiosize>%i</audiosize>\n", (int)flvmetadata.audiosize);
+	fprintf(fp, "<videosize>%i</videosize>\n", (int)flvmetadata.videosize);
+	fprintf(fp, "<filesize>%i</filesize>\n", (int)flvmetadata.filesize);
+
+	fprintf(fp, "<lasttimestamp>%.2f</lasttimestamp>\n", flvmetadata.lasttimestamp);
+	fprintf(fp, "<lastvideoframetimestamp>%.2f</lastvideoframetimestamp>\n", flvmetadata.lastvideoframetimestamp);
+	fprintf(fp, "<lastkeyframetimestamp>%.2f</lastkeyframetimestamp>\n", flvmetadata.lastkeyframetimestamp);
+	fprintf(fp, "<lastkeyframelocation>%i</lastkeyframelocation>\n", (int)flvmetadata.lastkeyframelocation);
+
+	fprintf(fp, "<keyframes>\n");
+	fprintf(fp, "<times>\n");
+
+	for(i = 0; i < flvmetadata.keyframes; ++i)
+		fprintf(fp, "<value id=\"%i\">%.2f</value>\n", i, flvmetadata.times[i]);
+
+	fprintf(fp, "</times>\n");
+	fprintf(fp, "<filepositions>\n");
+
+	for(i = 0; i < flvmetadata.keyframes; ++i)
+		fprintf(fp, "<value id=\"%i\">%i</value>\n", i, (int)flvmetadata.filepositions[i]);
+
+	fprintf(fp, "</filepositions>\n");
+	fprintf(fp, "</keyframes>\n");
+	fprintf(fp, "<duration>%.2f</duration>\n", flvmetadata.duration);
+	fprintf(fp, "</flv>\n");
+	fprintf(fp, "</fileset>\n");
+
+	return;
+}
+
 void writeFLVHeader(FILE *fp) {
 	char *t;
 	size_t size;
 	FLVFileHeader_t flvheader;
 
 	t = (char *)&flvheader;
-	bzero(t, sizeof(FLVFileHeader_t));
+	memset(t, 0, sizeof(FLVFileHeader_t));
 
 	flvheader.signature[0] = 'F';
 	flvheader.signature[1] = 'L';
@@ -689,7 +818,7 @@ void initFLVMetaData(const char *creator, int lastsecond, int lastkeyframe) {
 	char *t;
 
 	t = (char *)&flvmetadata;
-	bzero(t, sizeof(FLVMetaData_t));
+	memset(t, 0, sizeof(FLVMetaData_t));
 
 	flvmetadata.hasMetadata = 1;
 	flvmetadata.hasLastSecond = lastsecond;
@@ -716,7 +845,7 @@ size_t writeFLVMetaData(FILE *fp) {
 
 	// Alles auf 0 setzen
 	t = (char *)&flvtag;
-	bzero(t, sizeof(FLVTag_t));
+	memset(t, 0, sizeof(FLVTag_t));
 
 	// Tag Type
 	flvtag.type = FLV_SCRIPTDATAOBJECT;
@@ -871,6 +1000,8 @@ metadatapass:
 		goto metadatapass;
 	}
 
+	datasize += writeFLVScriptDataVariableArrayEnd(fp);
+
 	flvmetadata.metadatasize = datasize - sizeof(FLVTag_t);
 
 	datasize += writeFLVPreviousTagSize(fp, datasize);
@@ -878,8 +1009,9 @@ metadatapass:
 	return datasize;
 }
 
-size_t writeFLVLastSecond(FILE *fp) {
+size_t writeFLVLastSecond(FILE *fp, double timestamp) {
 	FLVTag_t flvtag;
+	int currenttimestamp;
 	size_t datasize = 0;
 	char *t;
 
@@ -887,11 +1019,18 @@ size_t writeFLVLastSecond(FILE *fp) {
 
 	// Alles auf 0 setzen
 	t = (char *)&flvtag;
-	bzero(t, sizeof(FLVTag_t));
+	memset(t, 0, sizeof(FLVTag_t));
 
 	// Tag Type
 	flvtag.type = FLV_SCRIPTDATAOBJECT;
 
+	// Timestamp
+	currenttimestamp = (int)(timestamp * 1000.0);
+	flvtag.timestamp_ex = ((currenttimestamp >> 24) & 0xff);
+	flvtag.timestamp[0] = ((currenttimestamp >> 16) & 0xff);
+	flvtag.timestamp[1] = ((currenttimestamp >> 8) & 0xff);
+	flvtag.timestamp[2] = (currenttimestamp & 0xff);
+
 	flvtag.datasize[0] = ((flvmetadata.onlastsecondlength >> 16) & 0xff);
 	flvtag.datasize[1] = ((flvmetadata.onlastsecondlength >> 8) & 0xff);
 	flvtag.datasize[2] = (flvmetadata.onlastsecondlength & 0xff);
@@ -923,7 +1062,7 @@ size_t writeFLVLastKeyframe(FILE *fp) {
 
 	// Alles auf 0 setzen
 	t = (char *)&flvtag;
-	bzero(t, sizeof(FLVTag_t));
+	memset(t, 0, sizeof(FLVTag_t));
 
 	// Tag Type
 	flvtag.type = FLV_SCRIPTDATAOBJECT;
@@ -1158,7 +1297,8 @@ void print_usage(void) {
 	fprintf(stderr, "\n");
 
 	fprintf(stderr, "SYNOPSIS\n");
-	fprintf(stderr, "\tyamdi -i input file -o output file [-c creator] [-h]\n");
+	fprintf(stderr, "\tyamdi -i input file [-x xml file | -o output file [-x xml file]]\n");
+	fprintf(stderr, "\t      [-c creator] [-l] [-h]\n");
 	fprintf(stderr, "\n");
 
 	fprintf(stderr, "DESCRIPTION\n");
@@ -1168,9 +1308,11 @@ void print_usage(void) {
 	fprintf(stderr, "\n");
 	fprintf(stderr, "\t-i\tThe source FLV file.\n");
 	fprintf(stderr, "\n");
-	fprintf(stderr, "\t-o\tThe resulting FLV file with the metatags. If the\n");
-	fprintf(stderr, "\t\toutput file is '-' the FLV file will be written to\n");
-	fprintf(stderr, "\t\tstdout.\n");
+	fprintf(stderr, "\t-o\tThe resulting FLV file with the metatags. If the file\n");
+	fprintf(stderr, "\t\tname is '-' the output will be written to stdout.\n");
+	fprintf(stderr, "\n");
+	fprintf(stderr, "\t-x\tAn XML file with the resulting metadata information. If the\n");
+	fprintf(stderr, "\t\toutput file is ommited, only metadata will be generated.\n");
 	fprintf(stderr, "\n");
 	fprintf(stderr, "\t-c\tA string that will be written into the creator tag.\n");
 	fprintf(stderr, "\n");
@@ -1180,7 +1322,7 @@ void print_usage(void) {
 	fprintf(stderr, "\n");
 
 	fprintf(stderr, "COPYRIGHT\n");
-	fprintf(stderr, "\t(c) 2007 Ingo Oppermann\n");
+	fprintf(stderr, "\t(c) 2008 Ingo Oppermann\n");
 	fprintf(stderr, "\n");
 	return;
 }
